import numpy as np
import pandas as pd
from pandas import DataFrame

from typing import List


def get_file_type(filename: str) -> str:
    return filename.split('.')[-1]


def read_txt_into_dataframe(file: str, skiprows: int = 1, **args) -> DataFrame:
    columns: List[str]
    with open(file, 'r') as f:
        if (delim := args.get("delimiter")) is None:
            delim = " "
        columns = f.readline().replace(delim, " ").replace("\n", "").split(" ")
    data = np.loadtxt(file, skiprows=skiprows, **args)
    return DataFrame(data=data, columns=columns)


def read_data_into_dataframe(file: str, **args) -> DataFrame:
    """tries to Convert the file into a DataFrame"""
    file_type = get_file_type(file)
    if file_type == "csv":
        return pd.read_csv(file, **args)
    elif file_type == "txt":
        return read_txt_into_dataframe(file, **args)
    else:
        raise ValueError(f"File type {file_type} not supported")



import pandas as pd

from dataclasses import dataclass, field
from typing import List

from .utils import transform_dataframe_to_latex_ready


@dataclass(repr=True)
class TexTabular:
    alignment: str
    data: pd.DataFrame = field(default_factory=pd.DataFrame, repr=False)
    caption: str = "default Caption"
    label: str = "default Label"
    caption_above: bool = True
    h_lines: List[int] | None = None
    filler: str = "--"
    booktabs: bool = False

    def __post_init__(self):
        if self.h_lines is None:
            self.h_lines = [1]
        if not self.data.empty:
            self.data = transform_dataframe_to_latex_ready(self.data)

    def add_data(self, data, **kwargs):
        self.data = transform_dataframe_to_latex_ready(data, **kwargs)

    def __data_to_str(self) -> str:
        result: str = ""
        data: list = [list(self.data.columns)]
        data.extend(list(i) for i in self.data.to_numpy())

        for (index, column) in enumerate(data):
            assert isinstance(self.h_lines, list)
            h_lines: int = len(list(filter(lambda x: x == index, self.h_lines)))
            if h_lines > 0:
                line_style = "hline"
                if index == 0 and self.booktabs:
                    line_style = "toprule"
                elif self.booktabs:
                    line_style = "midrule"

                result += "        " + f"\\{line_style}" * h_lines + "\n"
            result += "        "
            for elem in column:
                if str(elem) == "nan" or elem is None:
                    elem = self.filler
                result += str(elem) + " & "
            result = result.removesuffix(" & ")
            result += " \\\\\n"

        assert isinstance(self.h_lines, list)
        end_hlines: int = len(list(filter(lambda x: x == len(data), self.h_lines)))
        if end_hlines > 0:
            line_style = "hline"
            if self.booktabs:
                line_style = "bottomrule"

            result += "        " + f"\\{line_style}" * end_hlines + "\n"
        result = result.removesuffix("\n")
        return result

    def save(self, file_path: str, method: str = "w", positioning: str = "htbp"):
        result = f"\n\\begin{{table}}[{positioning}]\n    \\centering\n{str(self)}\n\\end{{table}}\n"
        with open(file_path, method) as file:
            file.write(result)

    def __str__(self):
        if self.data is None or self.data.empty:
            raise ValueError("Empty table cannot be printed. data attribute is empty.")
        caption_str = f"\\caption{{{self.caption}}}"
        label_str = f"\\label{{{self.label}}}"
        result: str = f"    \\begin{{tabular}}{{{self.alignment}}}\n"
        if self.caption_above:
            result = f"    {caption_str}\n    {label_str}\n{result}"
        result += f"{self.__data_to_str()}\n"
        result += f"    \\end{{tabular}}"
        if not self.caption_above:
            result += f"\n    {caption_str}\n    {label_str}"

        return result



"""Common physical constants in SI units."""

LIGHTSPEED: float = 299792458  # Lichtgeschwindigkeit
ELEMENTARY_CHARGE: float = 1.602176634e-19  # Elementarladung
VACUUM_PERMEABILITY: float = 1.25663706212e-6  # Magnetische Feldkonstante
VACUUM_PERMITTIVITY: float = 8.8541878128e-12  # Elektrische Feldkonstante
COULOMB_CONSTANT: float = 8.9875517922e+9  # Coulomb-Konstante
PLANCK_CONSTANT: float = 6.62607015e-34  # Plancksches Wirkungsquantum
GRAVITATIONAL_CONSTANT: float = 6.67430e-11  # Gravitationskonstante
PLANCK_MASS: float = 2.176434e-8  # Planck Masse
PLANCK_LENGTH: float = 1.616255e-35  # Planck Länge
PLANCK_TIME: float = 5.391247e-44  # Planck Zeit
PLANCK_TEMPERATURE: float = 1.486784e+32  # Planck Temperatur
BOLTZMANN_CONSTANT: float = 1.380649e-23  # Boltzmann Konstante
STEFAN_BOLTZMANN_CONSTANT: float = 5.670374419e-8  # Stefan Boltzmann Konstante
ELECTRON_MASS: float = 9.1093837015e-31  # Elektronenmasse
COMPTON_WAVELENGTH: float = 2.42631023867e-12  # Compton Wellenlänge
ELECTRON_RADIUS: float = 2.8179403262e-15  # klassischer Elektronenradius
RYDBERG_CONSTANT: float = 1.0973731568160e+7  # Rydberg konstante
RYDBERG_FREQUENCY: float = 3.2898419602508e15  # Rydberg Frequenz
AVOGADRO_CONSTANT: float = 6.02214076e23  # Avogadro Konstante
GAS_CONSTANT: float = AVOGADRO_CONSTANT * BOLTZMANN_CONSTANT  # Gaskonstante
# Vm0 = 22.41396954e-3 # Molares Volumen eines idealen Gases
ATOMIC_MASS: float = 1.66053906660e-27  # atomare Masseneinheit
FINE_STRUCTURE_CONSTANT: float = 7.297_352_5693e-3



import matplotlib.pyplot as plt
from matplotlib.figure import Figure, SubFigure
from mypy_extensions import VarArg
import numpy as np
from pandas import DataFrame, Series
import scienceplots
from scipy import odr
from scipy.optimize import curve_fit

from typing import List, Dict, Tuple, Iterable, Callable, TypeAlias, Optional, Any

from .fit_result import FitResult
from ..misc.dataframe import get_error_column_name
from ..misc.file_management import read_data_into_dataframe
from .parse import parse_function, replace_funcs, numerical, scalar, array

func_type: TypeAlias = Callable[[numerical, VarArg(scalar)], numerical]


class Fit:
    def __init__(self, meta_data: dict, data: DataFrame | None = None):
        self.meta_data = meta_data
        if data is None:
            self.file: str = meta_data["file"]
            if (args := meta_data.get("read_data_args")) is not None:
                self.data: DataFrame = read_data_into_dataframe(self.file, **args)
            else:
                self.data = read_data_into_dataframe(self.file)
        else:
            self.data = data

        self.result: Optional[dict[str, FitResult]] = None

        self.column_names = self.__get_column_names(meta_data.get("error_marker"))

    def __get_column_names(self, error_marker: List[str] | None = None) -> Dict[str, str]:
        if error_marker is None:
            error_marker = ["err", "error", "fehler", "Err", "Error", "Fehler"]
        names: Dict[str, str] = dict()
        names["x"] = self.meta_data["x"]
        names["y"] = self.meta_data["y"]
        x_error = get_error_column_name(self.data, names["x"], error_marker)
        y_error = get_error_column_name(self.data, names["y"], error_marker)
        if y_error is None:
            raise ValueError("Y Error values are missing")
        names["y error"] = y_error
        if x_error is not None:
            names["x error"] = x_error

        return names

    def plot(self, figure_id: Optional[int | str | Figure | SubFigure] = None):
        if (style := self.meta_data.get("figure_style")) is None:
            style = "science"
        plt.style.use(style)
        if (size := self.meta_data.get("figure_size")) is None:
            size = (6, 4.7)
        if (dpi := self.meta_data.get("figure_dpi")) is None:
            dpi = 120
        plt.figure(num = figure_id, figsize=size, dpi=dpi)
        x, y = self.column_names["x"], self.column_names["y"]
        sx, sy = self.column_names.get("x error"), self.column_names["y error"]
        if sx is None:
            sx = 0
        else:
            sx = self.data[sx]

        plotting_style = {
            "linestyle": "",
            "ms": 3,
            "marker": "o",
            "capsize": 2.5,
            "zorder": 100
        }

        if (more_style := self.meta_data.get("plotting_style")) is not None:
            plotting_style.update(more_style)

        plt.xlim(self.__get_xlim(self.meta_data))
        plt.errorbar(self.data[x], self.data[y], yerr=self.data[sy], xerr=sx, **plotting_style)

        if self.result is not None:
            for fit_name in self.result.keys():
                self.__plot_fit(fit_name, self.result[fit_name])

    def __plot_fit(self, name: str, fit_res: FitResult) -> None:
        meta: dict = self.meta_data["fits"][name]
        if (points := meta.get("fit_points")) is None:
            points = 150

        plotting_style = {
            "linestyle": "--",
            "zorder": 0,
        }

        if (more_style := meta.get("plotting_style")) is not None:
            plotting_style.update(more_style)

        x = np.linspace(*self.__get_xlim(meta), points)
        plt.plot(x, fit_res.eval(x), **plotting_style)

    def __get_xlim(self, meta: dict) -> Tuple[float, float]:
        x = self.column_names["x"]
        if (x_min := meta.get("x_min_limit")) is None:
            x_min = self.data[x].min()
        if (x_max := meta.get("x_max_limit")) is None:
            x_max = self.data[x].max()

        return x_min, x_max

    def run(self) -> dict[str, FitResult]:
        fits_meta: dict[str, dict] = self.meta_data["fits"]
        result: dict[str, FitResult] = dict()
        for fit_name in fits_meta.keys():
            result[fit_name] = self.__do_fit(fits_meta[fit_name])

        self.result = result

        return result

    def __do_fit(self, meta: dict) -> FitResult:
        if meta["fit_type"] == "optimize.curve_fit":
            return self.__do_optimize_curve_fit(meta)
        elif meta["fit_type"] == "odr":
            return self.__do_odr_fit(meta)
        else:
            raise ValueError(f"unknown fit type: {meta['fit_type']}")

    def __do_optimize_curve_fit(self, meta: dict) -> FitResult:
        function, params, p0, lim, arrays = self.__get_fitting_data(meta)
        x, y, y_error = arrays["x"], arrays["y"], arrays["sy"]
        assert isinstance(x, Series) and isinstance(y, Series) and isinstance(y_error, Series)

        if (b := meta.get("absolute_sigma")) is None:
            b = False
        if (check_finite := meta.get("check_finite")) is None:
            check_finite = False
        out: tuple = curve_fit(function, xdata=x, ydata=y, sigma=y_error, absolute_sigma=b,
                               check_finite=check_finite, p0=p0)
        popt, pcov = out
        fit_res = FitResult(function, params, popt, np.sqrt(pcov.diagonal()))
        fit_res.set_reduced_chi_squared(x, y, y_error)
        return fit_res

    def __do_odr_fit(self, meta: dict) -> FitResult:
        function, params, p0, lim, arrays = self.__get_fitting_data(meta)
        x, y, sy = arrays["x"], arrays["y"], arrays["sy"]
        sx = arrays.get("sx")
        assert isinstance(x, Series) and isinstance(y, Series) and isinstance(sy, Series)
        if (b := meta.get("absolute_sigma")) is None:
            b = False

        if b:
            odr_data = odr.RealData(x=x, y=y, sy=sy, sx=sx)
        else:
            wd = None
            if sx is not None:
                wd = 1 / sx ** 2
            odr_data = odr.Data(x=x, y=y, we=1 / sy ** 2, wd=wd)

        def odr_func(_b: list, _x: float | int) -> numerical:
            return function(_x, *_b)

        odr_model = odr.Model(odr_func)
        odr_odr = odr.ODR(odr_data, odr_model, p0)
        odr_out: odr.Output = odr_odr.run()
        fit_res = FitResult(function, params, odr_out.beta, odr_out.sd_beta)
        fit_res.set_reduced_chi_squared(x, y, sy)
        return fit_res

    def __get_fitting_data(self, meta: dict) -> Any:

        p0: List[float | int] = meta["start_parameters"]
        x_min, x_max = self.__get_xlim(meta)

        query = f"{self.column_names['x']} >= {x_min} and {self.column_names['x']} <= {x_max}"
        data = self.data.query(query)

        function, params = parse_function(replace_funcs(meta["function"]))
        x: Series = data[self.column_names["x"]]
        y: Series = data[self.column_names["y"]]
        sy: Series = data[self.column_names["y error"]]
        sx: Optional[Series] = None
        if (name := self.column_names.get("x error")) is not None:
            sx = data[name]

        fitting_data: Dict[str, array | Optional[array]] = {
            "x": x,
            "y": y,
            "sy": sy,
            "sx": sx,
        }

        lim: Tuple[float, float] = (x_min, x_max)

        return function, params, p0, lim, fitting_data



import numpy as np


# TODO change the reduced chi square function and translate the docstring to english.
def reduced_chi_square(f, x, y, y_err, params: list[float]) -> float:
    """Berechnet das X² pro Freiheitsgrad für eine Funktion f mit parametern <params>
    f: hat die Form f(params, __x)
    __x: ist ein Array der unabhängigen Variable
    y: ist ein Array der von __x abhängigen Variable
    y_err: Fehler von y, kann ein Array oder ein skalarer Wert sein"""

    try:
        iter(x)
        iter(y)
    except TypeError:
        print("__x and/ or y not an iterable")
        exit(-1)

    try:
        iter(y_err)
    except TypeError:
        y_err = [y_err] * len(x)

    x, y, y_err = np.array(x), np.array(y), np.array(y_err)

    chi_square = np.sum((y - f(params, x)) ** 2 / y_err ** 2)
    return chi_square / (len(x) - len(params))





from pandas import DataFrame

from typing import List


def get_error_column_name(data: DataFrame, column: str, error_marker: List[str]) -> str | None:
    """Get the columns name that is the uncertainty of the column argument. return None if not found.
    parameters
    ----------
    data: DataFrame
    The dataframe to analyze.
    column: str
    The name of the column which represents a series of values that have an uncertainty.
    error_marker: List[str]
    A list of words that indicate that a column represents a series of uncertainties.
    """
    columns = list(data.columns)
    if column not in columns:
        raise ValueError("column not in DataFrame")

    error_separator = ["_", "-", " ", ""]
    result: str | None = None
    for marker in error_marker:
        for separator in error_separator:
            if f"{column}{separator}{marker}" in columns:
                if result is None:
                    result = f"{column}{separator}{marker}"
                else:
                    text = f"""Error in DataFrame for the column {column} could not be determined because 
                    of ambivalence. Error marker are {error_marker}"""
                    raise ValueError(text)
    return result


import numpy as np
from typing import List


def roundup(x: float | int, r=2) -> float:
    """Rounds up a number to the decimal place r"""
    a = x*10**r
    a = np.ceil(a)
    a = a*10**(-r)

    if type(x) == float or type(x) == int or type(x) == np.float64:
        if a == 0:
            a = 10**(-r)
    else:
        try:                                           # rundet mehrdimensionale arrays
            for i, j in enumerate(a):
                for k, l in enumerate(j):
                    if i == 0:
                        i = 10**(-r)
        except:                                        # rundet eindimensionale arrays
            for i, j in enumerate(a):
                if i == 0:
                    i = 10**(-r)

    return np.around(a, r)


def roundup_two_significant_digits(x: float) -> float:
    """Rounds the given number up to 2 significant digits"""
    scientific: List[str] = "{:e}".format(x).split("e")
    value, exponent = float(scientific[0]), int(scientific[1])
    if value < 2:
        value = roundup(value, 1)
    else:
        value = roundup(value, 0)

    return float(f"{value}e{exponent}")



import pandas as pd
from mypy_extensions import VarArg
import numpy as np
from pandas import DataFrame

from dataclasses import dataclass, field
from typing import List, Dict, Callable, TypeAlias, Iterable

from ..mmath import NumWithError

scalar: TypeAlias = int | float
array: TypeAlias = np.ndarray | pd.Series
numerical: TypeAlias = scalar | array


@dataclass(repr=True)
class FitResult:
    function: Callable[[numerical, VarArg(scalar)], numerical] = field(repr=False)
    parameter_names: List[str]
    parameter_values: np.ndarray
    parameter_sigmas: np.ndarray
    reduced_chi_squared: float | None = field(default=None)

    def eval(self, x: numerical) -> float | np.ndarray:
        """Calculated values of the fitted function"""
        result = self.function(x, *self.parameter_values)
        if isinstance(result, Iterable):
            return np.array(result)
        else:
            return float(result)

    def as_dict(self, chi_square=False) -> dict[str, NumWithError]:
        result: dict = dict()
        for name, x, sigma in zip(self.parameter_names, self.parameter_values, self.parameter_sigmas):
            result[name] = NumWithError(x, sigma)

        if chi_square:
            result["reduced_chi_squared"] = self.reduced_chi_squared

        return result

    def set_reduced_chi_squared(self, x: array, y: array, sigma: array) -> float:
        chi_squared = ((self.eval(x) - y)**2/sigma**2).sum()
        result: float = chi_squared / (len(x) - len(self.parameter_names))
        self.reduced_chi_squared = result
        return result



import pandas as pd

from typing import List

from ..mmath.num_with_error import NumWithError
from ..misc.dataframe import get_error_column_name


def transform_dataframe_to_latex_ready(data, **kwargs):
    # initialize all keyword arguments
    if kwargs.get("error_marker") is None:
        kwargs["error_marker"] = ["err", "error", "fehler", "Err", "Error", "Fehler"]
    else:
        list(kwargs["error_marker"])
    if kwargs.get("columns") is None:
        kwargs["columns"] = {}
    else:
        dict(kwargs["columns"])
    if kwargs.get("ignore_rest") is None:
        kwargs["ignore_rest"] = False
    if not isinstance(kwargs["ignore_rest"], bool):
        raise TypeError("ignore_rest must be boolean")
    if kwargs.get("is_table_num") is None:
        kwargs["is_table_num"] = True
    if not isinstance(kwargs["is_table_num"], bool):
        raise TypeError("is_table must be a boolean")
    if kwargs.get("siunitx_column_option") is None:
        kwargs["siunitx_column_option"] = {}
    else:
        dict(kwargs["siunitx_column_option"])

    new_dataframe = data.copy()  # new dataframe for end result

    # look for columns with errors
    for column in data.columns:
        error = get_error_column_name(data, column, kwargs["error_marker"])
        if error is None:
            continue
        num_data = data[column]
        error_data = data[error]
        new_data = []
        option: str | None = kwargs["siunitx_column_option"].get(column)
        for x, err in zip(num_data, error_data):
            if kwargs["is_table_num"]:
                new_data.append(NumWithError(x, err).display_table_separate(option))
            else:
                new_data.append(NumWithError(x, err).display_separate(option))
        new_dataframe[column] = new_data
        new_dataframe.drop(error, axis=1, inplace=True)

    if kwargs["ignore_rest"]:
        columns = list(new_dataframe.columns)
        columns_to_ignore = [c for c in columns if c not in kwargs["columns"]]
        new_dataframe.drop(columns=columns_to_ignore, inplace=True)

    for col in list(new_dataframe.columns):
        option = kwargs["siunitx_column_option"].get(col)
        is_table_num: bool = kwargs["is_table_num"]
        new_dataframe[col] = [__display_num_value(i, option, is_table_num) if isinstance(i, (int, float)) else i
                              for i in new_dataframe[col]]

    new_dataframe.rename(columns=kwargs["columns"], inplace=True)

    return new_dataframe


def __display_num_value(x: int | float, option: str | None = None, is_table_num: bool = True) -> str | None:
    """takes a number and formats it to use with the siunitx package. Returns None if the value is not a number
    Example
    -------
    __display_num_value(2.3, sinuitx_option="round-mode=figures") -> \tablenum[round-mode=figures]{2.3}"""
    if str(x) == "nan":
        return None
    if option != "" and option is not None:
        option = f"[{option}]"
    else:
        option = ""

    macro = "tablenum"
    if not is_table_num:
        macro = "num"
    return f"\\{macro}{option}{{{x}}}"


if __name__ == "__main__":
    print(__display_num_value(2.23, option="round-mode=figures", is_table_num=False))



from mypy_extensions import VarArg
import nltk
import numpy as np
from pandas import Series

from typing import Callable, List, Tuple, TypeAlias, Iterable, Dict

scalar: TypeAlias = float | int
array: TypeAlias = Series | np.ndarray
numerical: TypeAlias = scalar | array
func_type: TypeAlias = Callable[[VarArg(str)], str]


def __gauss(*args: str) -> str:
    return f"({args[0]} * exp((-(x - {args[1]}) ** 2) / (2 * {args[2]} ** 2)))"


def __linear(*args: str) -> str:
    return f"({args[0]} * x + {args[1]})"


__ALPH_OPERATORS = {
    'arccos': 'np.arccos',
    'arccosh': 'np.arccosh',
    'arcsin': 'np.arcsin',
    'arcsinh': 'np.arcsinh',
    'arctan': 'np.arctan',
    'arctanh': 'np.arctanh',
    'cos': 'np.cos',
    'cosh': 'np.cosh',
    'e': 'np.e',
    'exp': 'np.exp',
    'log': 'np.log',
    'pi': 'np.pi',
    'sin': 'np.sin',
    'sinh': 'np.sinh',
    'tan': 'np.tan',
    'tanh': 'np.tanh',
}

__SYMBOL_OPERATORS = {
    '*': '*',
    '+': '+',
    '-': '-',
    '/': '/',
    '^': '**',
}

__FUNCS: Dict[str, Tuple[func_type, List[str]]] = {
    "gauss": (__gauss, ["a", "x0", "sigma"]),
    "linear": (__linear, ["m", "b"]),
}


def __get_unique_name(name: str, params: List[str]) -> str:
    if name not in params:
        return name
    i = 1
    while (out := f"{name}{i}") in params:
        i += 1
        continue

    return out


def __is_param(token: str) -> bool:
    answer = (token not in __ALPH_OPERATORS and token not in __FUNCS and token.isalnum() and
              not token.isnumeric() and token != "x")
    if answer and token[0].isnumeric():
        raise ValueError("param names cannot start with a number")
    return answer


def rename_parameters(_formula: str, rename: Dict[str, str]) -> str:
    tokens = nltk.tokenize.wordpunct_tokenize(_formula)
    for index, token in enumerate(tokens):
        for key in rename.keys():
            if key == token:
                tokens[index] = rename[key]

    return "".join(tokens)


def replace_funcs(_formula: str) -> str:
    tokens = nltk.tokenize.wordpunct_tokenize(_formula)
    params: List[str] = []
    for token in tokens:
        if __is_param(token) and token not in params:
            params.append(token)

    for index, token in enumerate(tokens):
        if token in __FUNCS:
            b = [__get_unique_name(i, params) for i in __FUNCS[token][1]]
            tokens[index] = __FUNCS[token][0](*b)
            params.extend(b)

    return "".join(tokens)


def parse_function(_formula: str) -> Tuple[Callable[[numerical, VarArg(float | int)], numerical], List[str]]:
    """Creates a python function out of a string that represents a mathematical formula.
    x is the variable and all the other arguments are passed as positional arguments to the
    generated callable with the form f(x: float, *params) -> float.
    Returns the callable and a list of all the arguments mentioned in _formula"""
    tokens = nltk.tokenize.wordpunct_tokenize(_formula)
    params: List[str] = []
    for index, token in enumerate(tokens):
        if token in __ALPH_OPERATORS:
            tokens[index] = __ALPH_OPERATORS[token]
        elif __is_param(token):
            params.append(token)

    def function(x: numerical, *_params: scalar) -> numerical:
        if isinstance(x, Iterable):
            return np.array([function(i, *_params) for i in x])
        else:
            try:
                float(x)
            except TypeError:
                raise TypeError(r"Argument x must be a number or an iterable.")
        nonlocal tokens, params
        _tokens = tokens.copy()
        for i, tok in enumerate(_tokens):
            if tok in params:
                param_index = params.index(tok)
                _tokens[i] = str(_params[param_index])
            if tok == "x":
                _tokens[i] = str(x)
        string = " ".join(_tokens)
        for op in __SYMBOL_OPERATORS.keys():
            string = string.replace(op, __SYMBOL_OPERATORS[op])
        return eval(string)

    return function, params



import pandas as pd

from dataclasses import dataclass
from typing import List


@dataclass
class TexTabular:
    r"""This classes purpose is to create a string which can be used to create a LaTeX tabular environment out of
    a pandas DataFrame. Because of a separate table class for more complex use, the tabular environment will not be
    automatically put inside a table environment, but it will be if printed to a file with the `save` method.

    the string will have output in the following form:
    \caption{#caption}
    \label{#label}
    \begin{tabular}{#alignment}
        #data
    \end{tabular}

    Parameters
    ----------
    alignment: str
    Determines the alignment of the table. For a table with 3 columns with centered
    outputs and columns separated by lines, the alignment would be "c|c|c".
    data: DataFrame, optional
    contains the data of the table. If the data is given at initialization, for every
    column the corresponding errors will be searched inside the DataFrame. If this is not desired, the
    data should be inserted after initialization. For more complex operations to the data like renaming
    the columns, the method `add_data` should be used. This parameter is only optional at initialization.
    caption: str, optional
    The caption of the table. Is set to "default Caption" by default.
    label: str, optional
    The label of the table. Is set to "default Label" by default.
    h_lines: List[int], optional
    contains the indices of positions, where horizontal lines should be added to
    the table. The index 0 represents a horizontal line above the first line. If an index is inserted to the
    table multiple times, multiples lines will be inserted at that position. Is set to [1] by default.
    filler: str, optional
    Will be used to fill spaces in the table where no data exists. is set to "--" by default.
    booktabs: bool
    If set to `True`, the hlines will be replaced with macros from the booktabs package
    (midrule, toprule, bottomrule).

    """
    alignment: str
    data: pd.DataFrame = ...
    caption: str = ...
    label: str = ...
    caption_above: bool = ...
    h_lines: List[int] | None = ...
    filler: str = ...
    booktabs: bool = ...

    def add_data(self, data: pd.DataFrame, **kwargs) -> None:
        r"""Iterates through the column names and tries to find the corresponding error column.
    If it is found, both columns will be put together into a new column with the numbers separated by a plusminus
    sign. Other numbers in the DataFrame will be put inside the tablenum macro for proper alignment inside the
    table.

    Parameters
    ----------
    data: DataFrame
    Data to be transformed into other data that can be made into a string to work with LaTeX tables.

    Keyword Parameters
    ------------------
    error_marker: List[str], optional
    Defines a list of all string suffixes that mark a column of uncertainties. The default value
    is ["err", "error", "fehler", "Err", "Error", "Fehler"].
    columns: Dict[str, str], optional
    If a Key corresponds to a column of the data, that column will be renamed to the given value.
    ignore_rest: bool, optional
    If set to true, every column of the data that is not a key in the `columns` keyword argument will
    be dropped in the resulting DataFrame. The default value is `False`.
    is_table_num: bool, optional
    If false, numbers will not be inserted in the tablenum macro but the num macro instead. Default value is `True`.
    siunitx_column_option: Dict[str, str], optional
    For every column that is a key in the given dictionary, the value is inserted as an option for the tablenum
    macro.
    """
    ...

    def __data_to_str(self) -> str:
        """Creates a string out of the data that can be inserted inside a LaTeX table"""
        ...

    def save(self, file_path: str, method: str = "w", positioning: str = "htbp") -> None:
        r"""Save the string inside a table environment to a file specified by `file_path`. The `method` argument
        corresponds to the same parameter of the builtin `open()` function.

        `position` determines the positioning behavior of the table.
        """
        ...




from typing import Tuple

from .rounding import roundup_two_significant_digits


def _round_values(x: float, x_error: float) -> Tuple[float, float, float]:
    """rounds up to 2 significant digits. functions only takes float values as arguments"""
    if not (isinstance(x, float) and isinstance(x_error, float)):
        raise ValueError("both x and x_error must be floating point numbers")
    if x_error <= 0:
        raise ValueError("x_error must be a number bigger than zero")

    x_error = roundup_two_significant_digits(x_error)
    value, exponent = "{:e}".format(x_error).split("e")

    rounding_decimal = - int(exponent)
    if value[2] != "0":
        rounding_decimal += 1

    x = round(x, rounding_decimal)

    return x, x_error, int(exponent)


class NumWithError:
    """inputs two numbers and rounds them appropriately, treating x_error as an uncertainty."""
    def __init__(self, x: float | int,  x_error: float | int) -> None:
        if isinstance(x, (float, int)) and isinstance(x_error, (float, int)):
            x, x_error = float(x), float(x_error)
            self.__x, self.__x_error, self.__exponent = _round_values(x, x_error)
        else:
            raise TypeError("x and x_error must be numbers")

    @property
    def x(self):
        return self.__x

    @property
    def x_error(self):
        return self.__x_error

    def get_values(self) -> Tuple[float, float]:
        return self.__x, self.__x_error

    def display_table_separate(self, option: str | None = None) -> str:
        r"""Creates a string of the form \tablenum{x +- x_error}"""
        if option is None:
            option = ''
        else:
            option = f"[{option}]"

        format_str = "{:}"
        if self.__exponent < 0:
            precision = str(-self.__exponent + 1)
            if "{:e}".format(self.__x_error)[2] == "0":
                precision = str(-self.__exponent)
            format_str = f"{{:.{precision}f}}"
        return f"\\tablenum{option}{{{format_str.format(self.x)} +- {format_str.format(self.x_error)}}}"

    def display_separate(self, option: str | None = None) -> str:
        r"""Same as display_table_separate but with \num instead of \tablenum."""
        output = self.display_table_separate(option)
        output = output.replace(r"\tablenum", r"\num")
        return output

    def __eq__(self, other):
        return self.__x == other.__x and self.__x_error == other.__x_error

    def __repr__(self):
        return f"NumWithError({self.__x}, {self.__x_error})"

    def __str__(self):
        return self.display_table_separate()



import unittest

from src.pymonke.mmath import NumWithError


class MyTestCase(unittest.TestCase):
    def test_num_with_error_eq(self):
        self.assertEqual(NumWithError(2.234, 0.123), NumWithError(2.23, 0.13))
        self.assertEqual(NumWithError(2.234, 0.14234), NumWithError(2.23, 0.15))
        self.assertEqual(NumWithError(2, 0.14234), NumWithError(2.00, 0.15))
        self.assertEqual(NumWithError(2, 0.16234), NumWithError(2.00, 0.17))
        self.assertEqual(NumWithError(2.234123, 0.16234), NumWithError(2.23, 0.17))
        self.assertEqual(NumWithError(2.234123, 0.19934), NumWithError(2.2, 0.2))
        self.assertEqual(NumWithError(0.9, 0.99), NumWithError(1, 1))
        self.assertEqual(NumWithError(10.8567, 0.99), NumWithError(11, 1))
        self.assertEqual(NumWithError(114.123, 10.89123), NumWithError(114, 11))
        self.assertEqual(NumWithError(0.8234, 10.914123), NumWithError(1, 11))
        self.assertEqual(NumWithError(0.08234, 10.914123), NumWithError(0, 11))
        self.assertEqual(NumWithError(0.8234, 103.814123), NumWithError(0, 110))
        self.assertEqual(NumWithError(0.8234, 110.0002), NumWithError(0, 120))
        self.assertEqual(NumWithError(1423.12341287, 103.814123), NumWithError(1420, 110))
        self.assertEqual(NumWithError(-2.3, 0.223), NumWithError(-2.3, 0.3))

    def test_num_with_error_values(self):
        self.assertEqual(NumWithError(2.234, 0.14234).get_values(), (2.23, 0.15))
        self.assertEqual(NumWithError(2, 0.14234).get_values(), (2.00, 0.15))
        self.assertEqual(NumWithError(2, 0.16234).get_values(), (2.00, 0.17))
        self.assertEqual(NumWithError(2.234123, 0.16234).get_values(), (2.23, 0.17))
        self.assertEqual(NumWithError(2.234123, 0.19934).get_values(), (2.2, 0.2))
        self.assertEqual(NumWithError(0.9, 0.99).get_values(), (1, 1))
        self.assertEqual(NumWithError(10.8567, 0.99).get_values(), (11, 1))
        self.assertEqual(NumWithError(114.123, 10.89123).get_values(), (114, 11))
        self.assertEqual(NumWithError(0.8234, 10.914123).get_values(), (1, 11))
        self.assertEqual(NumWithError(0.08234, 10.914123).get_values(), (0, 11))
        self.assertEqual(NumWithError(0.8234, 103.814123).get_values(), (0, 110))
        self.assertEqual(NumWithError(0.8234, 110.0002).get_values(), (0, 120))
        self.assertEqual(NumWithError(1423.12341287, 103.814123).get_values(), (1420, 110))

    def test_num_with_error_display_separate(self):
        self.assertEqual(NumWithError(2.234, 0.14234).display_separate(), r"\num{2.23 +- 0.15}")
        self.assertEqual(NumWithError(2, 0.14234).display_separate(), r"\num{2.00 +- 0.15}")
        self.assertEqual(NumWithError(2, 0.16234).display_separate(), r"\num{2.00 +- 0.17}")
        self.assertEqual(NumWithError(2.234123, 0.16234).display_separate(), r"\num{2.23 +- 0.17}")
        self.assertEqual(NumWithError(2.234123, 0.19934).display_separate(), r"\num{2.2 +- 0.2}")
        self.assertEqual(NumWithError(0.9, 0.99).display_separate(), r"\num{1.0 +- 1.0}")
        self.assertEqual(NumWithError(10.8567, 0.99).display_separate(), r"\num{11.0 +- 1.0}")
        self.assertEqual(NumWithError(114.123, 10.89123).display_separate(), r"\num{114.0 +- 11.0}")
        self.assertEqual(NumWithError(0.8234, 10.914123).display_separate(), r"\num{1.0 +- 11.0}")
        self.assertEqual(NumWithError(0.08234, 10.914123).display_separate(), r"\num{0.0 +- 11.0}")
        self.assertEqual(NumWithError(0.8234, 103.814123).display_separate(), r"\num{0.0 +- 110.0}")
        self.assertEqual(NumWithError(0.8234, 110.0002).display_separate(), r"\num{0.0 +- 120.0}")
        self.assertEqual(NumWithError(1423.12341287, 103.814123).display_separate(), r"\num{1420.0 +- 110.0}")
        self.assertEqual(NumWithError(0.000_000_000_712_312, 0.000_000_000_051_512_123).display_separate(),
                         r"\num{0.00000000071 +- 0.00000000006}")

        self.assertEqual(NumWithError(2.23e1, 14e1).display_separate("round-precision=1"),
                         r"\num[round-precision=1]{20.0 +- 140.0}")
        self.assertEqual(NumWithError(2.234, 0.14234).display_table_separate(), r"\tablenum{2.23 +- 0.15}")

    def test_num_with_error_raise(self):
        with self.assertRaises(TypeError):
            NumWithError(23, "not a number")
        with self.assertRaises(TypeError):
            NumWithError("not a number", 23)
        with self.assertRaises(TypeError):
            NumWithError([2.234, 2, "asd"], [0.14234, 0, 123])
        with self.assertRaises(TypeError):
            NumWithError([2, 3, 4], [0.14234, 0])
        with self.assertRaises(ValueError):
            NumWithError(2, 0)
        with self.assertRaises(ValueError):
            NumWithError(2, -0.123)
        with self.assertRaises(ValueError):
            NumWithError(-1.23, 0)
        with self.assertRaises(TypeError):
            NumWithError([2, 3], [-0.14234, 0, 0.123])


if __name__ == '__main__':
    unittest.main()



from pandas import DataFrame


def transform_dataframe_to_latex_ready(data: DataFrame, **kwargs) -> DataFrame:
    r"""Iterates through the column names and tries to find the corresponding error column.
    If it is found, both columns will be put together into a new column with the numbers separated by a plusminus
    sign. Other numbers in the DataFrame will be put inside the tablenum macro for proper alignment inside the
    table.

    Parameters
    ----------
    data: DataFrame
    Data to be transformed into other data that can be made into a string to work with LaTeX tables.

    Keyword Parameters
    ------------------
    error_marker: List[str], optional
    Defines a list of all string suffixes that mark a column of uncertainties. The default value
    is ["err", "error", "fehler", "Err", "Error", "Fehler"].
    columns: Dict[str, str], optional
    If a Key corresponds to a column of the data, that column will be renamed to the given value.
    ignore_rest: bool, optional
    If set to true, every column of the data that is not a key in the `columns` keyword argument will
    be dropped in the resulting DataFrame. The default value is `False`.
    is_table_num: bool, optional
    If false, numbers will not be inserted in the tablenum macro but the num macro instead. Default value is `True`.
    siunitx_column_option: Dict[str, str], optional
    For every column that is a key in the given dictionary, the value is inserted as an option for the tablenum
    macro.
    """
    ...



from dataclasses import dataclass, field
from typing import List

from .tex_tabular import TexTabular


@dataclass(repr=True)
class TexTable:
    r"""Class for generating a LaTeX table. The idea behind this class is to have the ability to
     create a table with multiple tabular environments, in the case where the user wants to have
     multiple tabular environments side by side.
     In the case where one tabular environment is enough, this class is not necessary to generate a LaTeX table.

     Parameters
     ----------
     positioning: str
     Is "htbp" by default and defines the position behavior of the table.
     tabular: List[TexTabular]
     Defines a list with tabular objects inside, which contain the data for the tabular environments.
     widths: List[float]
     Defines the space relative to the line width of the page that is given for the tabular environments.
     It is a list that should have the same length as the `tabular list.
     spacing: List[float | str]
     Defines the spacing between two tabular environments. It is a list which should be shorter by one
     than the `tabular` list. An element could be a string, which will be inserted verbosely, but in the case
     of a floating point number, it will be inserted as an argument to the `hspace` macro.
     """
    tabular: List[TexTabular]
    positioning: str = "htbp"
    widths: List[float] = field(default_factory=list)
    spacing: List[float | str] = field(default_factory=list)

    def save(self, file_path: str, method: str = "w") -> None:
        r"""Saves the LaTeX table to a file specified by `file_path`. The `method` argument is the same as
        for the builtin `open()` function."""
        with open(file_path, method) as file:
            file.write(str(self))

    def __str__(self) -> str:
        content = ""

        if len(self.tabular) == 0:
            raise ValueError("No tabular defined")
        elif len(self.tabular) == 1:
            content = self.tabular[0].__str__()
        else:
            if len(self.tabular) > len(self.widths):
                raise ValueError(r"you need to specify the width of each tabular")
            space: str = ""
            for (index, tab) in enumerate(self.tabular):
                text: str = space
                text += f"\n    \\parbox{{{self.widths[index]}\\linewidth}}{{\n        \\centering\n"
                text += str(tab)
                text += f"\n    }}"
                content += text
                try:
                    s = self.spacing[index]
                    if isinstance(s, float):
                        space = f"\\hspace{{{self.spacing[index]}cm}}"
                    elif isinstance(s, str):
                        space = s
                    else:
                        raise TypeError("the spacing parameter must be a float or a string")
                except IndexError:
                    space = r"\qquad"

        result = f"\\begin{{table}}[{self.positioning}]\n    \\centering\n{content}\n\\end{{table}}"
        return result



import unittest

from src.pymonke.mmath.rounding import roundup_two_significant_digits


class MyTestCase(unittest.TestCase):
    def test_roundup_two_significant_digits(self):
        def f(x):
            return roundup_two_significant_digits(x)

        self.assertEqual(f(0.123), 0.13)  # add assertion here
        self.assertEqual(f(0.1), 0.1)
        self.assertEqual(f(0.100123), 0.11)
        self.assertEqual(f(1234.2345), 1300)
        self.assertEqual(f(2234.2345), 3000)
        self.assertEqual(f(0.9), 0.9)
        self.assertEqual(f(0.8789789), 0.9)
        self.assertEqual(f(0.90234), 1.)


if __name__ == '__main__':
    unittest.main()



import unittest

import pandas as pd

from src.pymonke.latex.tex_table import TexTable
from src.pymonke.latex.tex_tabular import TexTabular

# TODO table_test.py not up to date

class MyTestCase(unittest.TestCase):
    def test_base_table(self):
        string = r"""\begin{table}[htbp]
    \centering
    \caption{Tabelle}
    \label{tab:tabelle}
    \begin{tabular}{c|c}
        \toprule
        Zahl1 & Zahl2 \\
        \midrule
        \tablenum{1} & \tablenum{1.0} \\
        \tablenum{2} & \tablenum{2.0} \\
        \tablenum{3} &   \\
        \bottomrule
    \end{tabular}
\end{table}"""
        data = pd.DataFrame(data={
            "Zahl1": [1, 2, 3],
            "Zahl2": [1, 2, None],
        })
        table = TexTable(tabular=[TexTabular(data=data, alignment="c|c", caption="Tabelle", label="tab:tabelle",
                                             h_lines=[0, 1, 4], filler=" ", booktabs=True)])
        self.assertEqual(string, str(table))

    def errors(self):

        with self.assertRaises(ValueError):
            table = TexTable(tabular=[TexTabular(), TexTabular()], widths=[0.34])
            print(table)


if __name__ == '__main__':
    unittest.main()
